@@ -3,19 +3,20 @@ import os
 def get_tracked_rst_files(repo=None):
     """Returns a list of .rst files that are under version control"""
 
-    #from mercurial.match import match
-    from mercurial.cmdutil import match
+    from mercurial.match import match
 
     if repo is None:
         repo = get_repo()
 
-    m = match(repo, ('*.rst',), {}, default='relglob')
+    m = match(repo, '.', ('*.rst',), {}, default='relglob')
     tracked = filter(lambda f: f in repo.dirstate, repo[None].walk(m))
     return [os.path.join(repo.root, f) for f in tracked]
 
 def get_untracked_rst_files(repo=None):
-    """Returns a list of all .rst files that are not currently tracked by
-    Mercurial"""
+    """
+    Returns a list of all .rst files that are not currently tracked by
+    Mercurial
+    """
 
     if repo is None:
         repo = get_repo()
