@@ -5,14 +5,16 @@
 
     The standard domain.
 
-    :copyright: Copyright 2007-2010 by the Sphinx team, see AUTHORS.
+    :copyright: Copyright 2007-2011 by the Sphinx team, see AUTHORS.
     :license: BSD, see LICENSE for details.
 """
 
 import re
+import unicodedata
 
 from docutils import nodes
 from docutils.parsers.rst import directives
+from docutils.statemachine import ViewList
 
 from sphinx import addnodes
 from sphinx.roles import XRefRole
@@ -20,7 +22,7 @@ from sphinx.locale import l_, _
 from sphinx.domains import Domain, ObjType
 from sphinx.directives import ObjectDescription
 from sphinx.util import ws_re
-from sphinx.util.nodes import make_refnode
+from sphinx.util.nodes import clean_astext, make_refnode
 from sphinx.util.compat import Directive
 
 
@@ -59,7 +61,7 @@ class GenericObject(ObjectDescription):
                 indextype = 'single'
                 indexentry = self.indextemplate % (name,)
             self.indexnode['entries'].append((indextype, indexentry,
-                                              targetname, targetname))
+                                              targetname, ''))
         self.env.domaindata['std']['objects'][self.objtype, name] = \
             self.env.docname, targetname
 
@@ -80,8 +82,8 @@ class EnvVarXRefRole(XRefRole):
         tgtid = 'index-%s' % env.new_serialno('index')
         indexnode = addnodes.index()
         indexnode['entries'] = [
-            ('single', varname, tgtid, varname),
-            ('single', _('environment variable; %s') % varname, tgtid, varname)
+            ('single', varname, tgtid, ''),
+            ('single', _('environment variable; %s') % varname, tgtid, '')
         ]
         targetnode = nodes.target('', '', ids=[tgtid])
         document.note_explicit_target(targetnode)
@@ -116,7 +118,7 @@ class Target(Directive):
                 indextype = indexentry[:colon].strip()
                 indexentry = indexentry[colon+1:].strip()
             inode = addnodes.index(entries=[(indextype, indexentry,
-                                             targetname, targetname)])
+                                             targetname, '')])
             ret.insert(0, inode)
         name = self.name
         if ':' in self.name:
@@ -159,7 +161,7 @@ class Cmdoption(ObjectDescription):
         self.indexnode['entries'].append(
             ('pair', _('%scommand line option; %s') %
              ((currprogram and currprogram + ' ' or ''), sig),
-             targetname, targetname))
+             targetname, ''))
         self.env.domaindata['std']['progoptions'][currprogram, name] = \
             self.env.docname, targetname
 
@@ -205,8 +207,8 @@ class OptionXRefRole(XRefRole):
 
 class Glossary(Directive):
     """
-    Directive to create a glossary with cross-reference targets
-    for :term: roles.
+    Directive to create a glossary with cross-reference targets for :term:
+    roles.
     """
 
     has_content = True
@@ -223,40 +225,103 @@ class Glossary(Directive):
         gloss_entries = env.temp_data.setdefault('gloss_entries', set())
         node = addnodes.glossary()
         node.document = self.state.document
-        self.state.nested_parse(self.content, self.content_offset, node)
 
-        # the content should be definition lists
-        dls = [child for child in node
-               if isinstance(child, nodes.definition_list)]
-        # now, extract definition terms to enable cross-reference creation
-        new_dl = nodes.definition_list()
-        new_dl['classes'].append('glossary')
+        # This directive implements a custom format of the reST definition list
+        # that allows multiple lines of terms before the definition.  This is
+        # easy to parse since we know that the contents of the glossary *must
+        # be* a definition list.
+
+        # first, collect single entries
+        entries = []
+        in_definition = True
+        was_empty = True
+        messages = []
+        for line, (source, lineno) in zip(self.content, self.content.items):
+            # empty line -> add to last definition
+            if not line:
+                if in_definition and entries:
+                    entries[-1][1].append('', source, lineno)
+                was_empty = True
+                continue
+            # unindented line -> a term
+            if line and not line[0].isspace():
+                # first term of definition
+                if in_definition:
+                    if not was_empty:
+                        messages.append(self.state.reporter.system_message(
+                            2, 'glossary term must be preceded by empty line',
+                            source=source, line=lineno))
+                    entries.append(([(line, source, lineno)], ViewList()))
+                    in_definition = False
+                # second term and following
+                else:
+                    if was_empty:
+                        messages.append(self.state.reporter.system_message(
+                            2, 'glossary terms must not be separated by empty '
+                            'lines', source=source, line=lineno))
+                    entries[-1][0].append((line, source, lineno))
+            else:
+                if not in_definition:
+                    # first line of definition, determines indentation
+                    in_definition = True
+                    indent_len = len(line) - len(line.lstrip())
+                entries[-1][1].append(line[indent_len:], source, lineno)
+            was_empty = False
+
+        # now, parse all the entries into a big definition list
         items = []
-        for dl in dls:
-            for li in dl.children:
-                if not li.children or not isinstance(li[0], nodes.term):
-                    continue
-                termtext = li.children[0].astext()
+        for terms, definition in entries:
+            termtexts = []
+            termnodes = []
+            system_messages = []
+            ids = []
+            for line, source, lineno in terms:
+                # parse the term with inline markup
+                res = self.state.inline_text(line, lineno)
+                system_messages.extend(res[1])
+
+                # get a text-only representation of the term and register it
+                # as a cross-reference target
+                tmp = nodes.paragraph('', '', *res[0])
+                termtext = tmp.astext()
                 new_id = 'term-' + nodes.make_id(termtext)
                 if new_id in gloss_entries:
                     new_id = 'term-' + str(len(gloss_entries))
                 gloss_entries.add(new_id)
-                li[0]['names'].append(new_id)
-                li[0]['ids'].append(new_id)
+                ids.append(new_id)
                 objects['term', termtext.lower()] = env.docname, new_id
+                termtexts.append(termtext)
                 # add an index entry too
                 indexnode = addnodes.index()
-                indexnode['entries'] = [('single', termtext, new_id, termtext)]
-                li.insert(0, indexnode)
-                items.append((termtext, li))
+                indexnode['entries'] = [('single', termtext, new_id, 'main')]
+                termnodes.append(indexnode)
+                termnodes.extend(res[0])
+                termnodes.append(addnodes.termsep())
+            # make a single "term" node with all the terms, separated by termsep
+            # nodes (remove the dangling trailing separator)
+            term = nodes.term('', '', *termnodes[:-1])
+            term['ids'].extend(ids)
+            term['names'].extend(ids)
+            term += system_messages
+
+            defnode = nodes.definition()
+            self.state.nested_parse(definition, definition.items[0][1], defnode)
+
+            items.append((termtexts,
+                          nodes.definition_list_item('', term, defnode)))
+
         if 'sorted' in self.options:
-            items.sort(key=lambda x: x[0].lower())
-        new_dl.extend(item[1] for item in items)
-        node.children = [new_dl]
-        return [node]
+            items.sort(key=lambda x:
+                       unicodedata.normalize('NFD', x[0][0].lower()))
+
+        dlist = nodes.definition_list()
+        dlist['classes'].append('glossary')
+        dlist.extend(item[1] for item in items)
+        node += dlist
+        return messages + [node]
 
 
-token_re = re.compile('`([a-z_][a-z0-9_]*)`')
+token_re = re.compile('`(\w+)`', re.U)
 
 def token_xrefs(text):
     retnodes = []
@@ -327,6 +392,8 @@ class StandardDomain(Domain):
     object_types = {
         'term': ObjType(l_('glossary term'), 'term', searchprio=-1),
         'token': ObjType(l_('grammar token'), 'token', searchprio=-1),
+        'label': ObjType(l_('reference label'), 'ref', 'keyword',
+                         searchprio=-1),
         'envvar': ObjType(l_('environment variable'), 'envvar'),
         'cmdoption': ObjType(l_('program option'), 'option'),
     }
@@ -340,15 +407,40 @@ class StandardDomain(Domain):
         'productionlist': ProductionList,
     }
     roles = {
-        'option': OptionXRefRole(innernodeclass=addnodes.literal_emphasis),
-        'envvar': EnvVarXRefRole(),
-        'token':  XRefRole(),
-        'term':   XRefRole(lowercase=True, innernodeclass=nodes.emphasis),
+        'option':  OptionXRefRole(innernodeclass=addnodes.literal_emphasis),
+        'envvar':  EnvVarXRefRole(),
+        # links to tokens in grammar productions
+        'token':   XRefRole(),
+        # links to terms in glossary
+        'term':    XRefRole(lowercase=True, innernodeclass=nodes.emphasis,
+                            warn_dangling=True),
+        # links to headings or arbitrary labels
+        'ref':     XRefRole(lowercase=True, innernodeclass=nodes.emphasis,
+                            warn_dangling=True),
+        # links to labels, without a different title
+        'keyword': XRefRole(warn_dangling=True),
     }
 
     initial_data = {
         'progoptions': {},  # (program, name) -> docname, labelid
         'objects': {},      # (type, name) -> docname, labelid
+        'labels': {         # labelname -> docname, labelid, sectionname
+            'genindex': ('genindex', '', l_('Index')),
+            'modindex': ('py-modindex', '', l_('Module Index')),
+            'search':   ('search', '', l_('Search Page')),
+        },
+        'anonlabels': {     # labelname -> docname, labelid
+            'genindex': ('genindex', ''),
+            'modindex': ('py-modindex', ''),
+            'search':   ('search', ''),
+        },
+    }
+
+    dangling_warnings = {
+        'term': 'term not in glossary: %(target)s',
+        'ref':  'undefined label: %(target)s (if the link has no caption '
+                'the label must precede a section header)',
+        'keyword': 'unknown keyword: %(target)s',
     }
 
     def clear_doc(self, docname):
@@ -358,32 +450,121 @@ class StandardDomain(Domain):
         for key, (fn, _) in self.data['objects'].items():
             if fn == docname:
                 del self.data['objects'][key]
+        for key, (fn, _, _) in self.data['labels'].items():
+            if fn == docname:
+                del self.data['labels'][key]
+        for key, (fn, _) in self.data['anonlabels'].items():
+            if fn == docname:
+                del self.data['anonlabels'][key]
+
+    def process_doc(self, env, docname, document):
+        labels, anonlabels = self.data['labels'], self.data['anonlabels']
+        for name, explicit in document.nametypes.iteritems():
+            if not explicit:
+                continue
+            labelid = document.nameids[name]
+            if labelid is None:
+                continue
+            node = document.ids[labelid]
+            if name.isdigit() or node.has_key('refuri') or \
+                   node.tagname.startswith('desc_'):
+                # ignore footnote labels, labels automatically generated from a
+                # link and object descriptions
+                continue
+            if name in labels:
+                env.warn_node('duplicate label %s, ' % name + 'other instance '
+                              'in ' + env.doc2path(labels[name][0]), node)
+            anonlabels[name] = docname, labelid
+            if node.tagname == 'section':
+                sectname = clean_astext(node[0]) # node[0] == title node
+            elif node.tagname == 'figure':
+                for n in node:
+                    if n.tagname == 'caption':
+                        sectname = clean_astext(n)
+                        break
+                else:
+                    continue
+            elif node.tagname == 'table':
+                for n in node:
+                    if n.tagname == 'title':
+                        sectname = clean_astext(n)
+                        break
+                else:
+                    continue
+            else:
+                # anonymous-only labels
+                continue
+            labels[name] = docname, labelid, sectname
 
     def resolve_xref(self, env, fromdocname, builder,
                      typ, target, node, contnode):
-        if typ == 'option':
+        if typ == 'ref':
+            if node['refexplicit']:
+                # reference to anonymous label; the reference uses
+                # the supplied link caption
+                docname, labelid = self.data['anonlabels'].get(target, ('',''))
+                sectname = node.astext()
+            else:
+                # reference to named label; the final node will
+                # contain the section name after the label
+                docname, labelid, sectname = self.data['labels'].get(target,
+                                                                     ('','',''))
+            if not docname:
+                return None
+            newnode = nodes.reference('', '', internal=True)
+            innernode = nodes.emphasis(sectname, sectname)
+            if docname == fromdocname:
+                newnode['refid'] = labelid
+            else:
+                # set more info in contnode; in case the
+                # get_relative_uri call raises NoUri,
+                # the builder will then have to resolve these
+                contnode = addnodes.pending_xref('')
+                contnode['refdocname'] = docname
+                contnode['refsectname'] = sectname
+                newnode['refuri'] = builder.get_relative_uri(
+                    fromdocname, docname)
+                if labelid:
+                    newnode['refuri'] += '#' + labelid
+            newnode.append(innernode)
+            return newnode
+        elif typ == 'keyword':
+            # keywords are oddballs: they are referenced by named labels
+            docname, labelid, _ = self.data['labels'].get(target, ('','',''))
+            if not docname:
+                return None
+            return make_refnode(builder, fromdocname, docname,
+                                labelid, contnode)
+        elif typ == 'option':
             progname = node['refprogram']
             docname, labelid = self.data['progoptions'].get((progname, target),
                                                             ('', ''))
             if not docname:
                 return None
-            else:
-                return make_refnode(builder, fromdocname, docname,
-                                    labelid, contnode)
+            return make_refnode(builder, fromdocname, docname,
+                                labelid, contnode)
         else:
-            docname, labelid = self.data['objects'].get((typ, target), ('', ''))
-            if not docname:
-                if typ == 'term':
-                    env.warn(node.get('refdoc', fromdocname),
-                             'term not in glossary: %s' % target, node.line)
-                return None
+            objtypes = self.objtypes_for_role(typ) or []
+            for objtype in objtypes:
+                if (objtype, target) in self.data['objects']:
+                    docname, labelid = self.data['objects'][objtype, target]
+                    break
             else:
-                return make_refnode(builder, fromdocname, docname,
-                                    labelid, contnode)
+                docname, labelid = '', ''
+            if not docname:
+                return None
+            return make_refnode(builder, fromdocname, docname,
+                                labelid, contnode)
 
     def get_objects(self):
         for (prog, option), info in self.data['progoptions'].iteritems():
-            yield (option, 'option', info[0], info[1], 1)
+            yield (option, option, 'option', info[0], info[1], 1)
         for (type, name), info in self.data['objects'].iteritems():
-            yield (name, type, info[0], info[1],
+            yield (name, name, type, info[0], info[1],
                    self.object_types[type].attrs['searchprio'])
+        for name, info in self.data['labels'].iteritems():
+            yield (name, info[2], 'label', info[0], info[1], -1)
+
+    def get_type_name(self, type, primary=False):
+        # never prepend "Default"
+        return type.lname
